/***********************************************************************

  c35sms: Main.h
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  1.0		Initial version

***********************************************************************/


#include <unistd.h>
#include <signal.h>
#include <string>
#include <stdexcept>
#include <fstream>
#include <cc++/slog.h>
#include <cc++/misc.h>
#include "Main.h"
#include "Utils.h"
#include "Server.h"
#include "Serial.h"


const char * ServerName = "c35sms"; /**< Server name */
const int ReplyOK = 200;            /**< Server OK reply */
const int ReplyError = 400;         /**< Server error reply */
bool sig_exit = false;              /**< Server exit flag */


/** Server configuration
 *
 * Configuration of the server.
 *
 */
typedef struct {
	InetAddress Bind;    /**< Server bind address */
	tpport_t Port;       /**< Server port */
	timeout_t Atimeout;  /**< Socket accept timeout */
	timeout_t IOtimeout; /**< Socket IO timeout */
	uid_t UID;           /**< Daemon UID */
	gid_t GID;           /**< Daemon GID */
	string Chroot;       /**< Root of chroot environment */
	string PIDfile;      /**< Path to PID file */
	bool Daemonize;      /**< Start as daemon */
	Addresses Addrs;     /**< Hosts allowed to connect to the server */
	string Device;       /**< Serial device with mobile phone */
	unsigned long Baud;  /**< Serial device baud rate */
	timeout_t Stimeout;  /**< Serial device timeout */
} server_config_t;



/** Configuration reader
 *
 * Class for reading the configuration file.
 *
 */
class ConfigReader : public Keydata {
	public:
		ConfigReader(server_config_t & config, Slog * syslog); /**< Constructor */
};


/** Constructor
 *
 * Read the configuration from
 * /etc/c35sms, section [general].
 *
 */
ConfigReader::ConfigReader(server_config_t & config, Slog * syslog)
	: Keydata((string("/") + string(ServerName) + string("/general")).c_str()) {
	
	const char * cval;
	
	try {
		cval = (*this)["bind"];
		if (cval)
			config.Bind = InetAddress(cval, (InetAddrValidator * ) NULL);
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid bind address, using default " << config.Bind << endl;
	}
	
	try {
		cval = (*this)["port"];
		if (cval) {
			long int port = intDecode(cval);
			if ((port < 0) || (port > 65535))
				throw invalid_argument("Illegal port number");
			config.Port = port;
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid port, using default " << config.Port << endl;
	}
	
	try {
		cval = (*this)["accept_timeout"];
		if (cval)
			config.Atimeout = intDecode(cval);
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid accept timeout, using default " << config.Atimeout << " ms" << endl;
	}
	
	try {
		cval = (*this)["io_timeout"];
		if (cval)
			config.IOtimeout = intDecode(cval);
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid I/O timeout, using default " << config.IOtimeout << " ms" << endl;
	}
	
	try {
		cval = (*this)["uid"];
		if (cval) {
			long int uid = intDecode(cval);
			if ((uid < 0) || (uid > 65535))
				throw invalid_argument("Illegal UID number");
			config.UID = uid;
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid UID, using default " << config.UID << endl;
	}
	
	try {
		cval = (*this)["gid"];
		if (cval) {
			long int gid = intDecode(cval);
			if ((gid < 0) || (gid > 65535))
				throw invalid_argument("Illegal GID number");
			config.UID = gid;
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid GID, using default " << config.GID << endl;
	}
	
	try {
		cval = (*this)["chroot"];
		if (cval)
			config.Chroot = cval;
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid chroot directory, using default " << config.Chroot << endl;
	}
	
	try {
		cval = (*this)["pidfile"];
		if (cval)
			config.PIDfile = cval;
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid PID file path, using default " << config.PIDfile << endl;
	}
	
	try {
		cval = (*this)["daemonize"];
		if (cval) {
			if ((strcmp(cval, "true") == 0) || (strcmp(cval, "yes") == 0) || (strcmp(cval, "1") == 0))
				config.Daemonize = true;
			else if ((strcmp(cval, "false") == 0) || (strcmp(cval, "no") == 0) || (strcmp(cval, "0") == 0))
				config.Daemonize = false;
			else
				throw invalid_argument("Invalid boolean argument");
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid daemonize flag, using default " << (config.Daemonize ? "true" : "false" ) << endl;
	}	

	try {
		const char * const * allow = getList("client");
		for (int i = 0; i < getCount("client"); i++) {
			InetAddress addr(allow[i], (InetAddrValidator * ) NULL);
			
			config.Addrs.push_back(addr);
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid client address, using build-in default list" << endl;
	}			
	
	try {
		cval = (*this)["device"];
		if (cval)
			config.Device = cval;
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid serial device, using default " << config.Device << endl;
	}
	
	try {
		cval = (*this)["serial_timeout"];
		if (cval)
			config.Stimeout = intDecode(cval);
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid serial timeout, using default " << config.Stimeout << " ms" << endl;
	}
	
	try {
		cval = (*this)["baud"];
		if (cval) {
			long int baud = intDecode(cval);
			if (baud < 0)
				throw invalid_argument("Baud rate must be positive");
			config.Baud = baud;
		}
	} catch(...) {
		if (syslog)
			slog(Slog::levelError) << "Configuration file error: invalid serial baud rate, using default " << config.Baud << " baud" << endl;
	}
}


/** Signal handler
 *
 * Handler for catching termination
 * signals and forcing the server
 * to exit cleanly.
 *
 * @param sig Catched signal.
 *
 */
void sighandler(int sig) {
	/* Force application exit by closing
	   all file descriptors the server
	   can sleep on (except stdin, stdout
	   and stderr) */
	for (int fd = getdtablesize(); fd > 2; fd--)
		close(fd);

	sig_exit = true;
}


/** Daemonize
 *
 * If the server should run as a daemon
 * then fork the current process.
 *
 * @return Return value of fork(2) or
 *         0 when not daemonized.
 *
 */
pid_t daemonize(const bool daemonize) {
	if (daemonize)
		return fork();
		
	return 0;
}


/** Detach from controlling terminal
 *
 * If the server should run as as daemon
 * then detach from current controlling terminal
 * and close all inherited files.
 *
 */
void detach(const bool daemonize) {
	if (daemonize) {
		/* Close all file descriptors inherited
		   from parent and reopen stdin, stdout
		   and stderr as /dev/null */
		int fd;
		for (fd = getdtablesize(); fd >= 0; fd--)
			close(fd);
		fd = open("/dev/null", O_RDWR);
		dup(fd);
		dup(fd);
		
		/* Detach from terminal */
		setsid();
	}
}


/** Main function
 *
 * Server main function. Command-line
 * and configuration file parsing, server
 * socket creating, detaching.
 *
 * @param argc Number of command-line arguments.
 * @param argv Command-line arguments (strings).
 *
 * @return Process return value, 0 on success.
 *
 */
int main(int argc, char * argv[]) {
	/* Setup syslog facility */
	slog(ServerName, Slog::classDaemon);
	
	/* Server configuration */
	server_config_t ServerConfig = { /* Default server configuration */
		InetAddress((InetAddrValidator * ) NULL),                  /* Bind any address (0.0.0.0) */
		9500,                                                      /* Port 9500 */
		1000,                                                      /* Accept timeout cycle 1 sec */
		60000,                                                     /* TCP socket I/O timeout 60 sec */
		65534,                                                     /* User ID: nobody */
		65534,                                                     /* Group ID: nobody */
		"/var/empty",                                              /* Chroot directory /var/empty */
		string("/var/run/") + string(ServerName) + string(".pid"), /* PID file /var/run/c35sms.pid */
		true,                                                      /* Fork detached daemon */
		Addresses(),                                               /* Empty list of allowed clients */
		"/dev/ttyS0",                                              /* COM1 port */
		19200,                                                     /* 19200 baud */
		10000                                                      /* Serial socket I/O timeout 10 sec */
	};
	ConfigReader cfgreader(ServerConfig, &slog);
	
	/* Daemonize */
	pid_t pid = daemonize(ServerConfig.Daemonize);
	if (pid == 0) { /* Daemon */
	
		/* Detach from terminal
		   if daemonized */
		detach(ServerConfig.Daemonize);
		
		try {
			/* Create server and serial objects */
			SMSSerial serial(ServerConfig.Device, ServerConfig.Stimeout, ServerConfig.Baud, &slog);
			SMSServer server(ServerConfig.Bind, ServerConfig.Port, ServerConfig.Addrs, &slog);
			
			/* Create PID file */
			fstream pidfile(ServerConfig.PIDfile.c_str(), ios::out);
			
			if (pidfile) {
				pidfile << getpid() << endl;
				pidfile.close();
			} else
				slog(Slog::levelError) << "Unable to create PID file " << ServerConfig.PIDfile << endl;
			
			/* Setup signal handling */
			signal(SIGHUP, SIG_IGN);
			signal(SIGCHLD, SIG_IGN);
			signal(SIGINT, &sighandler);
			signal(SIGQUIT, &sighandler);
			signal(SIGTERM, &sighandler);
		
			/* Security considerations */
			umask(027);
			chdir(ServerConfig.Chroot.c_str());
			chroot(ServerConfig.Chroot.c_str());
			setgid(ServerConfig.GID);
			setuid(ServerConfig.UID);
		
			/* Request loop */
			while (!sig_exit) {
				if (server.isPendingConnection(ServerConfig.Atimeout)) {
					try {
						TCPStream stream(server);
						
						stream.setTimeout(ServerConfig.IOtimeout);
						
						/* Get request */
						stream << ReplyOK << " " << ServerName << " ready" << endl;
						string request;
						if (getline(stream, request, '\n')) {
							if (request.substr(0, 5) == "RESET") { /* SMS reset request */
								try {
									serial.resetMobile();
									stream << ReplyOK << " Mobile phone reset" << endl;
								} catch (runtime_error & rt) {
									stream << ReplyError << " " << rt.what() << endl;
								}
							} else if (request.substr(0, 4) == "SEND") { /* SMS send request */
								bool ready = true;
								SMSNumbers numbers;
								if (request.length() > 6) {
									StringTokenizer tokens(request.substr(5).c_str(), " \t", true, true);
								
									try {
										for (StringTokenizer::iterator it = tokens.begin(); it != tokens.end(); ++it) {
											string str_it = *it;
											SMSNumber number(str_it);
											
											numbers.push_back(number);
										}
									} catch (invalid_argument & inv) {
										stream << ReplyError << " " << inv.what() << endl;
										ready = false;
									}
								} else {
									stream << ReplyError << " Empty GSM numbers list" << endl;
									ready = false;
								}
								
								if (ready) {
									stream << ReplyOK << " Ready to get message body" << endl;
									
									/* Process message body */
									bool first = true;
									bool ok = false;
									string body;
									while (!sig_exit) {
										
										/* Read the input line by line
										   and concatenate the message body
										   (the end of the body is denoted
										   by a single period on the beginning
										   of a line; otherwise the single period
										   is ignored) */
										string line;
										if (getline(stream, line, '\n')) {
											
											/* Body termination */
											if (line == ".") {
												ok = true;
												break;
											}
										
											/* Ignore period */
											if ((line.length() > 0) && (line[0] == '.'))
												line.erase(0, 1);
											
											/* Insert newline from previous
											   line */
											if (first)
												first = false;
											else
												body += '\n';
											
											/* Concatenate message body */
											body += line;
										} else
											break;
									}
									if (ok) {
										try {
											/* Do send SMS */
											SMSNumbers::size_type mout = serial.sendSMS(numbers, body);
											if (mout == numbers.size())
												stream << ReplyOK << " Message sent OK" << endl;
											else
												stream << ReplyError << " Message send error (" << mout << " messages sent)" << endl;
										} catch (runtime_error & rt) {
											stream << ReplyError << " " << rt.what() << endl;
										} catch (invalid_argument & inv) {
											stream << ReplyError << " " << inv.what() << endl;
										}
									} else
										stream << ReplyError << " Message body not terminated" << endl;
								}
							
							} else {
								stream << ReplyError << " Unknown request" << endl;
								slog(Slog::levelInfo) << "Unknown request" << endl;
							}
						}
						
						tpport_t pport;
						InetHostAddress paddr = stream.getPeer(&pport);
						
						slog(Slog::levelInfo) << "Connection from " << paddr << ":" << pport << " closed" << endl;
					} catch (Socket * socket) {
						
						/* Connection problems inside server
						   process are not treated fatal */
						const char * errorString = socket->getErrorString();
						Socket::Error errorNumber = socket->getErrorNumber();
			
						if (errorString)
							slog(Slog::levelInfo) << "Socket failed (" << errorString << ")" << endl;
						else
							slog(Slog::levelInfo) << "Socket failed (error code " << errorNumber << ")" << endl;
						
					}
				}
			}
		} catch (Socket * socket) {
			
			/* Socket error */
			const char * errorString = socket->getErrorString();
			Socket::Error errorNumber = socket->getErrorNumber();
			
			if (errorString)
				slog(Slog::levelError) << "Socket failed (" << errorString << ")" << endl;
			else {
				switch (errorNumber) {
					case Socket::errBindingFailed:
						slog(Slog::levelError) << "Socket failed (requested port " << ServerConfig.Port << " is busy)" << endl;
						break;
					case Socket::errResourceFailure:
						slog(Slog::levelError) << "Socket failed (resource failure)" << endl;
						break;
					default:
						slog(Slog::levelError) << "Socket failed (error code " << errorNumber << ")" << endl;
				}
			}
			
			return -1;
		} catch (Serial * serial) {
			
			/* Serial error */
			const char * errorString = serial->getErrorString();
			Serial::Error errorNumber = serial->getErrorNumber();
			
			if (errorString)
				slog(Slog::levelError) << "Serial device failed (" << errorString << ")" << endl;
			else
				slog(Slog::levelError) << "Serial device failed (error code " << errorNumber << ")" << endl;
			
			return -1;
		} catch (runtime_error & rt) {
			slog(Slog::levelError) << rt.what() << endl;
			return -1;
		}
		
		slog(Slog::levelInfo) << "Daemon exiting" << endl;
		return 0;
		
	} else if (pid > 0) { /* Parent process */
		slog(Slog::levelInfo) << "Running as daemon with PID " << pid << endl;
		return 0;
	}
	
	slog(Slog::levelError) << "Unable to fork" << endl;
	return -1;
}
