/***********************************************************************

  c35sms: Serial.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  1.0		Initial version

***********************************************************************/


#include <stdexcept>
#include "Serial.h"
#include "Utils.h"


const char AT_prefix[] = "AT";     /**< AT commands prefix */
const char AT_param_delim[] = "="; /**< AT parameters delimiter */
const string OK = "OK";            /**< AT response OK */

const char cmd_init[] = "Z";         /**< Initialize mobile phone to defaults */
const char cmd_charset[] = "+CSCS";  /**< Set charset */
const char cmd_sms_send[] = "+CMGS"; /**< Send SMS (without storing to a slot) */

const char number_type_mask = 0x70;     /**< GSM number type mask */
const char number_type_internat = 0x10; /**< Internatinal number bits */


/** Default constructor, always throws invalid_argument
 *
 * Because empty SMS number is not a valid number,
 * this constructor always throws invalid_argument.
 *
 */
SMSNumber::SMSNumber() {
	throw invalid_argument("Empty SMS number is not valid");
}


/** Conversion constructor
 *
 * Converts the SMS number from a string,
 * checking for validity (throws invalid_argument
 * on invalid format).
 *
 * @param number SMS number to use.
 *
 */
SMSNumber::SMSNumber(string & number) {
	assign(number);
}


/** Conversion assign operator
 *
 * Converts the SMS number from a string,
 * checking for validity (throws invalid_argument
 * on invalid format).
 *
 * @param number SMS number to use.
 *
 */
SMSNumber & SMSNumber::operator=(string & number) {
	assign(number);
	return *this;
}


/** Conversion operator
 *
 * Get (valid) SMS number as string.
 *
 * @return SMS number as string.
 *
 */
SMSNumber::operator const string & () const {
	return value;
}


/** Get GSM type of the number
 *
 * @return GSM number type.
 *
 */
char SMSNumber::numType() const {
	if ((value.length() == 0) || (value[0] != '+'))
		return 0x81;
	
	return 0x91;
}


/** Assign new value, check for correctness
 *
 * Throws invalid_argument for invalid GSM
 * numbers.
 *
 * @param number String to parse.
 *
 */
void SMSNumber::assign(string & number) {
	int state = 0;
	int digits = 0;

	value = "";
	for (string::iterator i = number.begin(); i != number.end(); i++) {
		switch (*i) {
			case '+':
				switch (state) {
					case 0: /* '+' as first character */
						value = "+";
						state = 1;
						break;
					case 1:
						throw invalid_argument("Misplaced '+' in " + number);
				}
				break;
			case '0'...'9':
				switch (state) {
					case 0:
						state = 1;
					case 1:
						value += *i;
						digits++;
						break;
				}
				break;
			default:
				throw invalid_argument("Misplaced character in " + number);
		}
	}
	
	if ((digits < 1) || (digits > 20))
		throw invalid_argument("Invalid number of digits in " + number);
}


/** Constructor
 *
 * Open serial device.
 *
 * @param device  Device node.
 * @param timeout I/O timeout.
 * @param baud    Baud rate.
 *
 */
SMSSerial::SMSSerial(const string & device, const timeout_t timeout, const unsigned long baud, Slog * logger)
	: TTYStream(device.c_str(), timeout) {
	
	syslog = logger;
	
	/* Set serial line parameters */
	setFlowControl(flowHard);
	setParity(parityNone);
	setCharBits(8);
	setSpeed(baud);
	setStopBits(1);
	interactive(true);
}


/** Send SMS message
 *
 * Throws invalid_argument if the message is too
 * long or contains unsupported characters.
 *
 * @param numbers List of numbers to send
 *                the message to.
 * @param msg     Message body.
 *
 * @return Number of successfully sent messages.
 *
 */
SMSNumbers::size_type SMSSerial::sendSMS(SMSNumbers & numbers, string & msg) {
	/* Initialize */
	resetMobile();

	SMSNumbers::size_type cnt = 0;
	string::size_type msglen = 0;
	
	/* Convert message to GSM encoding */
	string encmsg = gsmEncode(msg, msglen);
	if (encmsg.length() > 160)
		throw invalid_argument("The message size is limited to 160 GSM characters");

	for (SMSNumbers::iterator it = numbers.begin(); it != numbers.end(); it++) {
		
		char number_type = it->numType();
		string number;
		if ((number_type & number_type_mask) == number_type_internat)
			number = ((string) (*it)).substr(1);
		else
			number = *it;
			
		string pdu; /* Magic tag */
		
		pdu += hexEncode(0);
		pdu += hexEncode(17);
		pdu += hexEncode(0);
		pdu += hexEncode(number.length());
		pdu += hexEncode(number_type);
		
		for (string::size_type i = 0; i < number.length(); i += 2) {
			if (i + 1 < number.length())
				pdu += number[i + 1];
			else
				pdu += 'F';
			pdu += number[i];
		}
		
		pdu += hexEncode(0);
		pdu += hexEncode(0);
		pdu += hexEncode(0xff);
		pdu += hexEncode(msglen);
		pdu += encmsg;
		
		/* Send AT command, wait for prompt */
		sendCommand(false, cmd_sms_send, decEncode((pdu.length() - 2) / 2));
		string res = getPrompt();
		if (res != ">")
			return cnt;
		
		/* Send message metadata and data */
		(*this) << pdu << "\x1a";
		getResponse(true); /* PDU echo */
		res = getResponse(true); /* Message reference */
		if (syslog)
			(*syslog)(Slog::levelInfo) << "Sent SMS message with reference " << res << endl;
		res = getResponse(true);
		if (res != OK)
			return cnt;
		
		cnt++;
	}

	return cnt;
}


/** Reset mobile phone
 *
 * Reset mobile phone to initial
 * configuration.
 *
 */
void SMSSerial::resetMobile() {
	/* Cleanup any pending input */
	(*this) << "\x1a";
	while (isPending(pendingInput, 0))
		getResponse(true);
	cmds.clear();

	/* Initialize device */
	if (sendCommand(true, cmd_init) != OK)
		throw runtime_error("Mobile init failed");

	/* Set default charset to GSM */
	if (sendCommand(true, cmd_charset, "\"GSM\"") != OK)
		throw runtime_error("Mobile set charset command failed");
}


/** Send AT command
 *
 * Send AT command and get response.
 *
 * @param ack    Wait for standard response.
 * @param cmd    Command (w/o AT).
 * @param params Optional parameters.
 *
 * @return Response.
 *
 */
string SMSSerial::sendCommand(const bool ack, const char * cmd, const string & params) {
	if (cmd == NULL)
		throw invalid_argument("NULL command");
	
	/* Construct command */	
	string send = string(AT_prefix) + string(cmd);
	if (params.length() > 0)
		send += string(AT_param_delim) + params;
	send += '\r';
	
	/* Push to the commands list
	   and send to the port */
	cmds.push_back(send);
	(*this) << send;
	
	return getResponse(ack);
}


/** Get command response
 *
 * Get the response of a command. Ignores
 * empty lines and collects all previous
 * command echoes. If 'ack' is set, then
 * awaits the last response in form
 * 'OK' or 'ERROR' on a separate line.
 *
 * @param ack Wait for standard response.
 *
 * @return Response.
 *
 */
string SMSSerial::getResponse(const bool ack) {
	string response;
	
	while ((!cmds.empty()) || (response.length() == 0)) {
		/* Read line from serial device */
		if (!std::getline(*this, response, '\n'))
			return "";
		
		/* Strip any trailing \r */
		if ((response.length() > 0) && (response[response.length() - 1] == '\r'))
			response.erase(response.length() - 1, 1);
		
		/* Collect command echoes */
		if (!cmds.empty()) {
			const string cmd = cmds.front();
			if (response == cmd) {
				cmds.pop_front();
				if (ack)
					response = "";
				continue;
			}
		}
	}
	
	return response;
}


/** Get prompt
 *
 * Read the input delimited by spaces
 * and get the prompt.
 *
 * @return Prompt.
 *
 */
string SMSSerial::getPrompt() {
	string response;
	
	while (response.length() == 0) {
		if (!std::getline(*this, response, ' '))
			return "";
		
		/* Strip any trailing \r */
		if ((response.length() > 0) && (response[response.length() - 1] == '\r'))
			response.erase(response.length() - 1, 1);
	}
	
	return response;
}
