/***********************************************************************

  c35sms: Utils.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  1.0		Initial version

***********************************************************************/


#include <stdint.h>
#include <langinfo.h>
#include <iconv.h>
#include <sstream>
#include <stdexcept>
#include "Utils.h"


/** UTF-32 Unicode charset description
 *
 */
#ifdef WORDS_BIGENDIAN
#	define INTERNAL_CHARSET "UCS-4BE"
#else
#	define INTERNAL_CHARSET "UCS-4LE"
#endif


typedef uint32_t ucs4char_t; /**< UTF-32 Unicode character */


/** Convert integer into two hexadecimal digits
 *
 * Values in range 0 .. 255 are represented
 * as two-digit hexadecimal string '00' .. 'FF'.
 *
 * @param value Integer value.
 *
 * @return Representation of value in hexadecimal
 *         two digit form (i.e. '9A').
 *
 */
string hexEncode(const int value) {
	ostringstream stream;
	
	stream.width(2);
	stream.fill('0');
	stream << hex << uppercase << (value & 255);
	return stream.str();
}


/** Convert integer into decimal representation
 *
 * @param value Integer value.
 *
 * @return Value in decimal string representation.
 *
 */
string decEncode(const int value) {
	ostringstream stream;
	
	stream << value;
	return stream.str();
}


/** Convert string to long int
 *
 * If the string doesn't represent a valid
 * integer value, invalid_argument is thrown.
 *
 * @param value String containing representation
 *              of an integer.
 *
 * @return Integer value.
 *
 */
long int intDecode(const string & decimal) {
	char * err;
	
	long int val = strtol(decimal.c_str(), &err, 0);
	if ((err == NULL) || (*err != (char) 0))
		throw invalid_argument("Not an integer");
	
	return val;
}


/** Convert string to GSM encoding
 *
 * Converts message string from local
 * 8b character encoding to GSM encoding.
 *
 * @param msg Message string in local encoding.
 * @param len Length of the message (in characters,
 *            not in octets).
 *
 * @return Message in GSM encoding.
 *
 */
string gsmEncode(string & msg, string::size_type & len) {
	/* Convert via Unicode */
	iconv_t cd = iconv_open(INTERNAL_CHARSET, nl_langinfo(CODESET));
	if (cd == ((iconv_t) -1))
		throw runtime_error("Unable to initialize iconv");
	
	string str;
	try {		
		
		for (string::iterator it = msg.begin(); it != msg.end(); it++) {
			char ch = *it;
			ucs4char_t uch = 0;
			
			size_t ch_size = sizeof(ch);
			size_t uch_size = sizeof(uch);
			
			char * ch_ptr = &ch;
			char * uch_ptr = (char * ) &uch;
			
			/* Convert from local to UTF-32 */
			size_t ret = iconv(cd, &ch_ptr, &ch_size, &uch_ptr, &uch_size);
			if (ret == ((size_t) -1))
				throw runtime_error("iconv() error");
				
			/* Convert from UTF-32 to GSM */
			switch (uch) {
				case 0x0040: /* commercial at (@) */
					str += (char) 0x00;
					break;
				case 0x00a3: /* pound sign */
					str += 0x01;
					break;
				case 0x0024: /* dollar sign ($) */
					str += 0x02;
					break;
				case 0x00a5: /* yen sign */
					str += 0x03;
					break;
				case 0x00e8: /* small 'e' with grave */
					str += 0x04;
					break;
				case 0x00e9: /* small 'e' with acute */
					str += 0x05;
					break;
				case 0x00f9: /* small 'u' with grave */
					str += 0x06;
					break;
				case 0x00ec: /* small 'i' with grave */
					str += 0x07;
					break;
				case 0x00f2: /* small 'o' with grave */
					str += 0x08;
					break;
				case 0x00c7: /* capital 'C' with cedilla */
				case 0x00e7: /* small 'c' with cedilla */
					str += 0x09;
					break;
				case 0x00d8: /* capital 'O' with stroke */
					str += 0x0b;
					break;
				case 0x00f8: /* small 'o' with stroke */
					str += 0x0c;
					break;
				case 0x00c5: /* capital 'A' with ring above */
					str += 0x0e;
					break;
				case 0x00e5: /* small 'a' with ring above */
					str += 0x0f;
					break;
				case 0x0394: /* capital delta */
					str += 0x10;
					break;
				case 0x005f: /* underscore / low line */
					str += 0x11;
					break;
				case 0x03a6: /* capital phi */
					str += 0x12;
					break;
				case 0x0393: /* capital gamma */
					str += 0x13;
					break;
				case 0x039b: /* capital lambda */
					str += 0x14;
					break;
				case 0x03a9: /* capital omega */
					str += 0x15;
					break;
				case 0x03a0: /* capital pi */
					str += 0x16;
					break;
				case 0x03a8: /* capital psi */
					str += 0x17;
					break;
				case 0x03a3: /* capital sigma */
					str += 0x18;
					break;
				case 0x0398: /* capital theta */
					str += 0x19;
					break;
				case 0x039e: /* capital xi */
					str += 0x1a;
					break;
				case 0x000c: /* form feed */
					str += 0x1b;
					str += 0x0a;
					break;
				case 0x005e: /* circumflex accent */
					str += 0x1b;
					str += 0x14;
					break;
				case 0x007b: /* left curly bracket */
					str += 0x1b;
					str += 0x28;
					break;
				case 0x007d: /* right curly bracket */
					str += 0x1b;
					str += 0x29;
					break;
				case 0x005c: /* reverse solidus (back slash) */
					str += 0x1b;
					str += 0x2f;
					break;
				case 0x005b: /* left square bracket */
					str += 0x1b;
					str += 0x3c;
					break;
				case 0x007e: /* tilde */
					str += 0x1b;
					str += 0x3d;
					break;
				case 0x005d: /* right square bracket */
					str += 0x1b;
					str += 0x3e;
					break;
				case 0x007c: /* vertical line */
					str += 0x1b;
					str += 0x40;
					break;
				case 0x20ac: /* Euro sign */
					str += 0x1b;
					str += 0x65;
					break;
				case 0x00c6: /* capital 'AE' */
					str += 0x1c;
					break;
				case 0x00e6: /* small 'ae' */
					str += 0x1d;
					break;
				case 0x00df: /* small sharp s (german) */
					str += 0x1e;
					break;
				case 0x00c9: /* capital 'E' with acute */
					str += 0x1f;
					break;
				case 0x00a4: /* currency sign */
				case 0x00b0: /* degree sign (not offical, preference of author) */
					str += 0x24;
					break;
				case 0x00a1: /* inverted exclamation mark */
					str += 0x40;
					break;
				case 0x0391: /* capital alpha */
					str += 0x41;
					break;
				case 0x0392: /* capital beta */
					str += 0x42;
					break;
				case 0x0395: /* capital epsilon */
					str += 0x45;
					break;
				case 0x0397: /* capital eta */
					str += 0x48;
					break;
				case 0x0399: /* capital iota */
					str += 0x49;
					break;
				case 0x039a: /* capital kappa */
					str += 0x4b;
					break;
				case 0x039c: /* capital mu */
					str += 0x4d;
					break;
				case 0x039d: /* capital nu */
					str += 0x4e;
					break;
				case 0x039f: /* capital omicron */
					str += 0x4f;
					break;
				case 0x03a1: /* capital rho */
					str += 0x50;
					break;
				case 0x03a4: /* capital tau */
					str += 0x54;
					break;
				case 0x03a5: /* capital upsilon */
					str += 0x55;
					break;
				case 0x03a7: /* capital chi */
					str += 0x58;
					break;
				case 0x0396: /* capital zeta */
					str += 0x5a;
					break;
				case 0x00c4: /* capital 'A' with diaeresis */
					str += 0x5b;
					break;
				case 0x00d6: /* capital 'O' with diaeresis */
					str += 0x5c;
					break;
				case 0x00d1: /* capital 'N' with tilde */
					str += 0x5d;
					break;
				case 0x00dc: /* capital 'U' with diaeresis */
					str += 0x5e;
					break;
				case 0x00a7: /* section sign */
					str += 0x5f;
					break;
				case 0x00bf: /* inverted question mark */
					str += 0x60;
					break;
				case 0x00e4: /* small 'a' with diaeresis */
					str += 0x7b;
					break;
				case 0x00f6: /* small 'o' with diaeresis */
					str += 0x7c;
					break;
				case 0x00f1: /* small 'n' with tilde */
					str += 0x7d;
					break;
				case 0x00fc: /* small 'u' with diaeresis */
					str += 0x7e;
					break;
				case 0x00e0: /* small 'a' with grave */
					str += 0x7f;
					break;
				default:
					if ((uch < 0x0080) && (
						(uch == 0x000A) ||
						(uch == 0x000D) ||
						((0x0020 <= uch) && (uch <= 0x0023)) || 
						((0x0025 <= uch) && (uch <= 0x003f)) || 
						((0x0041 <= uch) && (uch <= 0x005a)) || 
						((0x0061 <= uch) && (uch <= 0x007a)))) {
						str += uch;
					} else
						throw invalid_argument("Invalid GSM character");
			}
		}
	} catch (...) {
		/* Cleanup */
		iconv_close(cd);
		throw;
	}
	
	/* 7 bit GSM octet encoding */
	len = str.length();
	string output;
	unsigned int state = 0;
	
	for (string::iterator it = str.begin(); it != str.end(); it++) {
		if (state != 7) {
			/* The character is 7 bit long
			   (shift depends on state) */
			state %= 8;
			string::iterator next = it + 1;
			char first;
			if (next != str.end())
				first = (((*next) & 0x7F) << (7 - state)) & 0xFF;
			else
				first = 0;
			char last = ((*it) & 0x7F) >> state;
			
			output += hexEncode(first | last);
		}
		state++;
	}
	
	iconv_close(cd);
	return output;
}
